//
//   @file    fastnlo_utils.cxx
//
//
//   @author D. Britzger, K. Rabbertz
//
//
// Get preprocessor defines from autotools setup
#include "amconfig.h"

// NNLO bridge includes
#include "nnlo_common.h"
#include "nnlo_utils.h"

// Require fastNLO
#ifdef HAVE_FASTNLO

// System includes
#include <algorithm>
#include <cfloat>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <sys/stat.h>
#include <unistd.h>
#include <utility>

// fastNLO includes
#include "fastnlo_utils.h"
//#include "fastnlotk/fastNLOCreate.h"

//-----------------------------------------------------------------------------
//
// --- --- --- Access external functions --- --- ---
//
//-----------------------------------------------------------------------------
/// these should be in some common header !
extern "C" double alphaspdf_(const double& Q);
extern "C" void   evolvepdf_(const double& x, const double& Q, double* xf);
extern "C" bool   getunitphase_();

static double _myfillwgt = 0;


//-----------------------------------------------------------------------------
//
// --- --- --- fastNLO_nnlo --- --- ---
//
//-----------------------------------------------------------------------------

// _____________________________________________________________________ //
void nnlo::init_fastnlo( const int& id, const std::string& gridname, const int& nbins, const double* bins ) {
   //!<
   //!< Remember init variables for later real initialization at first fill call, see InitFastNLO below
   //!<
   if ( id >= (int)fnloUtils::fInits.size() ) {
      fnloUtils::ftable.resize(id+1);
      fnloUtils::fIsInclJet.resize(id+1);
      fnloUtils::fIs820GeV.resize(id+1);
      fnloUtils::fIs137.resize(id+1);
      fnloUtils::ftablename.resize(id+1);
      fnloUtils::fInits.resize(id+1);
   }
   fnloUtils::fInits[id] = fnloUtils::fnloInit{int(id), std::string(gridname), int(nbins), bins };
   fnloUtils::ftable[id] = NULL;
}

// _____________________________________________________________________ //
void nnlo::fill_fastnlo( const int& id, const double& obs, const double* wt, const double* wtref ) {
   //!<
   //!< Fill fastNLO table; at first call do table initialization
   //!<

   // { // --- debugging
   //    say::SetGlobalVerbosity(say::INFO);
   //    double x1 = parfrac_.x1;
   //    double x2 = parfrac_.x2;
   //    double Q2 = mur2_hook(1);
   //    double xfx[13] = {0};
   //    evolvepdf_(x1,sqrt(Q2),xfx);
   //    double as = alphaspdf_(sqrt(Q2));

   //    std::cout<<"id="<<id<<"\tfnloProc="<<fnloUtils::sp[currentprocess_.iproc]
   //          <<"\tx1="<<x1
   //          <<"\tx2="<<x2
   //          <<"\tQ2="<<Q2
   //          <<"\tobs="<<obs
   //          <<"\tgluon-x1="<<xfx[6]
   //          <<"\tas="<<as
   //          <<"\twt="<<wt[0]<<std::endl;

   // }

   // --- Do some counting & print out
   static unsigned int nfills = 0;
   static unsigned int nmess  = 0;
   static unsigned int nfmod  = 1;
   nfills++;
   if ( nmess < 10 && nfills % nfmod == 0 ) {
      nmess++;
      std::cout << "[nnlo::fill_fastnlo]: Call no. " << nfills << " of fastnlo fill routine." << std::endl;
   } else if ( nmess == 10 ) {
      nmess = 0;
      nfmod *= 10;
   }

   // --- DEPRECATED, was from pre COMPOSITE histo times
   // static const int obs_id_ptj1 = nnlo::GetNLegs() >= 1 ? obs_id_hook("ptj1") : -1;
   // static const int obs_id_ptj2 = nnlo::GetNLegs() >= 2 ? obs_id_hook("ptj2") : -1;
   // static const int obs_id_ptj3 = nnlo::GetNLegs() >= 3 ? obs_id_hook("ptj3") : -1;
   // static const int obs_id_ptj4 = nnlo::GetNLegs() >= 4 ? obs_id_hook("ptj4") : -1;

   // --- init fastNLO
   if ( fnloUtils::ftable[id]==NULL ) fnloUtils::InitFastNLO(id, wt);
   // TODO What is the following line for? Test of table creation?)
   if ( fnloUtils::ftable[id]==NULL ) {
     std::cout << "[nnlo::fill_fastnlo] TODO: Table not initialised for id = " << id << std::endl;
     std::cout << "[nnlo::fill_fastnlo] Should not happen, but ignored for now."  << std::endl;
     return;
   }
   // Check UNIT_PHASE setting of NNLOJET
   // If yes ==> weights nonsense, only phase space counts for grid warmup
   // InitFastNLO stops production runs with UNIT_PHASE on
   static const bool unitphase = getunitphase_();

   // --- Some sanity checks
   // Check process index
   static unsigned int npidx = 0;
   static unsigned int npmss = 0;
   static unsigned int npmod = 1;
   if ( fnloUtils::sp[currentprocess_.iproc] == -1 ) {
      npidx++;
      if ( npmss < 10 && npidx % npmod == 0 ) {
         npmss++;
         std::cerr << "[nnlo::fill_fastnlo] Error! Negative process index iproc = "<<currentprocess_.iproc<<", ProcId = "<<fnloUtils::sp[currentprocess_.iproc]<<std::endl;
      } else if ( npmss == 10 ) {
         npmss = 0;
         npmod *= 10;
      }
   }
   // Check sum of all absolute(!) weights
   // maxscl is a global var in nnlo_common.h and must be equal to maxscl_hook info from NNLOJET (checked in nnlo::init)
   double wtsum = 0.;
   for ( int is = 0; is < maxscl; is++ ) {
     wtsum += std::abs(wt[is]);
   }
   // Tiny weights and not grid warmup ==> nothing to do
   static unsigned int nzwgt = 0;
   static unsigned int nzmss = 0;
   static unsigned int nzmod = 1;
   // if ( fabs(wt[0]) < DBL_MIN && !unitphase ) { // WRONG!!!
   if ( wtsum < DBL_MIN && !unitphase ) {
      nzwgt++;
      if ( nzmss < 10 && nzwgt % nzmod == 0 ) {
         nzmss++;
         std::cout << "[nnlo::fill_fastnlo] " << nzwgt << "th occurrence of return because of tiny scale weight sum = " << wtsum << std::endl;
      } else if ( nzmss == 10 ) {
         nzmss = 0;
         nzmod *= 10;
      }
      return;
   }

   // --- Keep track of what we have calculated
   fnloUtils::RememberContribution();

   // --- Calculate weights and x-values
   // maxscl is a global var in nnlo_common.h and must be equal to maxscl_hook info from NNLOJET (checked in nnlo::init)
   // Furthermore it is assumed that unfilled scale variations are zero!
   double __w2[maxscl] = {0};
   double __x1=1, __x2=1;
   for ( int is = 0 ; is<maxscl ;is++ ) {
     if ( std::abs(wt[is]) > DBL_MIN ) {
         __w2[is] = wt[is];
         // DB DIS special cases TODO
         if ( fnloUtils::fIs137[id] ) { // ZEUS is using 1./137
            __w2[is] *= 1./eweakz_.amzZ/eweakz_.amzZ/137./137.;
         }
         if ( fnloUtils::fIs820GeV[id] ) { //
            __w2[is] *= 820./920.;
         }
         nnlo::GetFillWeights(__w2[is],__x1,__x2,is);
         // x values < 0 should not happen!
         // Check copied to nnlo::GetFillWeights ==> weight set to zero, x values to unity
         // static unsigned int nxneg = 0;
         // static unsigned int nxmss = 0;
         // static unsigned int nxmod = 1;
         // if ( (__x1 < 0) || ( !nnlo::IsDIS() && __x2 < 0 ) ) {
         //   nxneg++;
         //   if ( nxmss < 10 && nxneg % nxmod == 0 ) {
         //     nxmss++;
         //     std::cerr<<"[nnlo::fill_fastnlo] ERROR count "<<nxneg<<" ! x1 or x2 from GetFillWeights smaller than 0! x1="<<__x1<<", x2="<<__x2<<std::endl;
         //   } else if ( nxmss == 10 ) {
         //     nxmss = 0;
         //     nxmod *= 10;
         //   }
         // }
         if ( unitphase ) __w2[is]=0.0;
     } else {
       __w2[is]=0.0;
     }
   }

   // --- Calculate reference weight for cross check with NNLOJET; ignore during warmup
   const bool DoTest = false;
   if ( DoTest ) {
      if ( !fnloUtils::ftable[id]->GetIsWarmup() && !unitphase ) {
         fnloUtils::CalculateReferenceWeight(__w2[0],__x1,__x2);
      }
   }

   // --- Pass variables to fastNLO
   // Collect multiple results in case of DIS &| inclusive jets
   std::vector<double> vObs;
   // --- DEPRECATED, was from pre COMPOSITE histo times
   // if ( nnlo::IsDIS() && fnloUtils::fIsInclJet[id] ) { // TODO Why AND with nnlo::DIS ???
   //    vObs.resize(nnlo::GetNLegs());
   //    if ( vObs.size()>=1 ) vObs[0] = obs_val_hook(obs_id_ptj1);
   //    if ( vObs.size()>=2 ) vObs[1] = obs_val_hook(obs_id_ptj2);
   //    if ( vObs.size()>=3 ) vObs[2] = obs_val_hook(obs_id_ptj3);
   //    if ( vObs.size()>=4 ) vObs[3] = obs_val_hook(obs_id_ptj4);
   //    for ( auto& iv : vObs ) {
   //       if ( std::abs(iv) < DBL_MIN ) iv=0.0;
   //    }
   // } else {
      vObs.push_back(obs);
   // }

   bool isflex = fnloUtils::ftable[id]->GetIsFlexibleScale();
   for ( unsigned int iObs = 0 ; iObs<vObs.size() ; iObs++ ) {
      double thisobs = vObs[iObs];
      // TODO What for?
      if ( (thisobs == 0) || (std::abs(thisobs) < DBL_MIN) ) {
         std::cout<<"Filling: IsIncl="<<fnloUtils::fIsInclJet[id]<<"\tobs="<<thisobs<<"\tcontinue"<<std::endl;
         continue;
      }
      //      else if ( std::abs(thisobs) < DBL_MIN ) continue;
      // TODO flex-scale arithmetics; check with DB
      // Flexible-scale tables
      if ( isflex ) {
         double mf  = sqrt(muf2_hook(1));
         double mr  = sqrt(mur2_hook(1));
         std::vector<double> wscl = fnloUtils::SclIndepWgt(__w2,mf,mr);
         double s1 = 0;
         double s2 = 0;
         // Enable use of ptjet as scale for inclusive jets
         if ( fnloUtils::fIsInclJet[id] ) {
            // DIS: s1=Q2, s2=ptjet; pp: s1=ptjet, s2=mffromruncard
            s1 = nnlo::IsDIS() ? mf : thisobs;
            s2 = nnlo::IsDIS() ? thisobs : mf;
         }
         // Use mf and mu settings from NNLOJET run card to define scale1 and scale2 for flex-scale table
         // Recall to use proper asymmetric mur,muf settings for comparison when evaluating fastNLO grid!
         else {
            static int id_obs_mf = std::abs(mf/(int)(mf+0.5)-1) < 1.e-10 ? (int)(mf+0.5) : -1; // fabs(mf-round(mf)) < 1.e-10
            static int id_obs_mr = std::abs(mr/(int)(mr+0.5)-1) < 1.e-10 ? (int)(mr+0.5) : -1;
            if ( id_obs_mf != -1 ) mf = obs_val_hook(id_obs_mf);
            if ( id_obs_mr != -1 ) mr = obs_val_hook(id_obs_mr);
            s1 = nnlo::IsDIS() ? mf : mr; // muf: scale1 (pp: scale2)
            s2 = nnlo::IsDIS() ? mr : mf; // mur: scale2 (pp: scale1)
            if ( s1==0 || s2==0 ) {
               std::cout<<"WARNING! scale is zero (this happens if a 2-jet observable is taken as scale for a 1-jet observable): s1="<<s1<<"\ts2="<<s2<<"\tobs="<<thisobs<<"\tIsIncl="<<fnloUtils::fIsInclJet[id]<<"\tid="<<id<<"\t"<<fnloUtils::ftable[id]->GetFilename()<<std::endl;
            }
         }
         //std::cout<<"Filling: IsIncl="<<fnloUtils::fIsInclJet[id]<<"\tobs="<<thisobs<<"\ts1="<<s1<<"\ts2="<<s2<<std::endl;
         fnloUtils::ftable[id]->fEvent.SetProcessId(fnloUtils::sp[currentprocess_.iproc]);
         fnloUtils::ftable[id]->fEvent.SetX1(__x1);
         if ( __x1 < 0 ) std::cerr<<"fastnlo_utils. ERROR. x1 smaller than 0!! x="<<__x1<<std::endl;
         fnloUtils::ftable[id]->fEvent.SetX2(__x2);
         fnloUtils::ftable[id]->fEvent.SetSigma(wtref[0]);
         if ( fnloUtils::fIs820GeV[id] ) { //
            double x820 = __x1 * 920./820.;
            fnloUtils::ftable[id]->fEvent.SetX1(x820);
            if ( x820 > 1 ) continue;
         }
         fnloUtils::ftable[id]->fScenario.SetObservable0(thisobs);
         fnloUtils::ftable[id]->fScenario.SetObsScale1(s1); // in DIS, this must be Q2. In pp we take 'mur' from the runcard
         fnloUtils::ftable[id]->fScenario.SetObsScale2(s2); // in DIS, this is mur from the runcard. In pp we take 'muf' from the runcard
         // --- fill event within fastNLO
         fnloUtils::ftable[id]->fEvent.SetWeight_MuIndependent(wscl[0]);
         if ( nnlo::GetOrder() >= 1 ) {
            if ( nnlo::IsDIS() ){
               // TODO --> DB Is the following necessary?
               //fnloUtils::ftable[id]->fEvent.SetWeight_MuIndependent(wscl[0] + log(s1)*(wscl[1]+wscl[2]));  // DIS needs log(mu^2/q2) factors for ln(mur) and ln(muf) terms
               fnloUtils::ftable[id]->fEvent.SetWeight_MuIndependent(wscl[0]);  // DIS: new convention: no log(q2) terms any longer
               fnloUtils::ftable[id]->fEvent.SetWeight_log_mur(wscl[1]);
               fnloUtils::ftable[id]->fEvent.SetWeight_log_muf(wscl[2]);
               fnloUtils::ftable[id]->fEvent.SetWeight_log_murr(wscl[3]); // DIS: no log(q2) factor here!
            } else {
               fnloUtils::ftable[id]->fEvent.SetWeight_log_mur(wscl[1]);
               fnloUtils::ftable[id]->fEvent.SetWeight_log_muf(wscl[2]);
               fnloUtils::ftable[id]->fEvent.SetWeight_log_murr(wscl[3]);
            }
            if ( nnlo::GetOrder() >= 2 ) {
               fnloUtils::ftable[id]->fEvent.SetWeight_log_muff(wscl[4]);
               fnloUtils::ftable[id]->fEvent.SetWeight_log_murf(wscl[5]);
            }
         }
         fnloUtils::ftable[id]->Fill();
      }
      // Fixed-scale tables
      else {
        int nScl = fnloUtils::ftable[id]->GetScaleVarMax();
        //         int nScl = nnlo::GetOrder()>0 ? 3 : 1; // fastNLO convention
        if ( (unitphase || fnloUtils::ftable[id]->GetIsWarmup()) && nScl != 1 ) {
           std::cerr<<"[nnlo::fill_fastnlo] ERROR! No. of scale variations should always be one at warmup. nScl = "<<nScl<<" Aborted."<<std::endl;
           exit(13);
         }
         for ( int is = 0 ; is<nScl ; is++ ) {
            double fillweight = __w2[is];
            // --- Ignore meaningless weights during warmup
            if (fnloUtils::ftable[id]->GetIsWarmup()) fillweight = 1.0;
            if (std::abs(fillweight) > DBL_MIN || unitphase) {
               double mur2 = mur2_hook(is+1);
               if ( fnloUtils::fIsInclJet[id] ) { // TODO
                  mur2 = thisobs*thisobs * mur2_hook(is+1)/mur2_hook(1); // TODO: Scale factors ... where are the multiple entries???
               }
               fnloUtils::ftable[id]->fEvent.SetProcessId(fnloUtils::sp[currentprocess_.iproc]);
               fnloUtils::ftable[id]->fEvent.SetX1(__x1);
               fnloUtils::ftable[id]->fEvent.SetX2(__x2);
               fnloUtils::ftable[id]->fScenario.SetObservable0(thisobs);
               fnloUtils::ftable[id]->fScenario.SetObsScale1(sqrt(mur2));
               // --- Fill event into fastNLO table
               fnloUtils::ftable[id]->fEvent.SetWeight(fillweight);
               fnloUtils::ftable[id]->Fill(is);
            }
         }
      }
   }
   return;

}

// _____________________________________________________________________ //
void nnlo::fill_ref_fastnlo( const int& id, const double& obs, const double* wt ) {
   // Do nothing for the moment
   // test weights for real radiation
   // if ( wt[0]!=0 ) std::cout<<" wt/ref="<<_myfillwgt/wt[0]<<"\t\ttest wt="<<wt[0]<<"\twtfill="<<_myfillwgt<<std::endl;
   // if ( wt[0]!=0 ) std::cout<<"                  "<<"\t\ttest wt="<<wt[0]<<"\twtfill="<<_myfillwgt<<std::endl;

   // KR: Use C++ macro constants like DBL_MIN instead
   //   if (  fabs(wt[0])> TINY && _myfillwgt!=0 && fabs(1-_myfillwgt/wt[0])>1.e-5 ) {
   if (  std::abs(wt[0])> DBL_MIN && _myfillwgt!=0 && std::abs(1-_myfillwgt/wt[0])>1.e-5 ) {
      std::cout<<"ERROR !! Your fill weight is not correctly calculated!"<<std::endl;
      std::cout<<"  wt/ref="<<_myfillwgt/wt[0]<<"\t\ttest wt="<<wt[0]<<"\twtfill="<<_myfillwgt<<std::endl;
      std::cout<<"  currentprocess_.iproc:   "<<currentprocess_.iproc<< std::endl;
      std::cout<<"  xregions_.xreg1      :   "<<xregions_.xreg1<<"\t1./xr1="<<1./xregions_.xreg1<< std::endl;
      std::cout<<"  xregions_.xreg2      :   "<<xregions_.xreg2<<"\t1./xr1="<<1./xregions_.xreg2<< std::endl;
      std::cout<<"  parfrac_.x1          :   "<<parfrac_.x1<<"\t1./x1="<<1./parfrac_.x1<<"\t(1-x1)="<<1-parfrac_.x1<<"\t1./(1-x1)="<<1./(1-parfrac_.x1)<< std::endl;
      std::cout<<"  parfrac_.x2          :   "<<parfrac_.x2<<"\t1./x2="<<1./parfrac_.x2<<"\t(1-x2)="<<1-parfrac_.x2<<"\t1./(1-x2)="<<1./(1-parfrac_.x2)<< std::endl;
      //cout<<"  ndim                 :   "<<ndimcurrent_.ndim<< std::endl;
      std::cout<<"  nv                   :   "<<ndimcurrent_.nv<< std::endl;
      std::cout<<"  iproc                :   "<<currentprocess_.iproc<< std::endl;
      std::cout<<"  igx                  :   "<<gridix_.igx<< std::endl;
      std::cout<<"  colflag              :   "<<colourflag_.colflag<< std::endl;
      std::cout<<"  xcoljac              :   "<<xcoljac_.xcoljac<<"\t\t1./xcoljac="<<1./xcoljac_.xcoljac<< std::endl;
      std::cout<<"  jacobian_.jflag      :   "<<jacobian_.jflag<< std::endl;

      static int ii=0;
      if ( ii++ >50 ) {
         std::cout <<" ende test weight and 'x' value for real radiation. use gluon channels only! " <<std::endl;
         exit(1);
      }
    }

}

// _____________________________________________________________________ //
void nnlo::term_fastnlo( const int& id, int nweights ) {
   //!<
   //!< Terminate fastNLO; normalise and write table to disk
   //!<

   // TODO What does the following line mean?
   if ( vegasiterweight_.it != runinfo_.itmax2 ) return;
   std::cout << "[nnlo::term_fastnlo] Write fastNLO table id="<<id<<" with nweights " <<nweights << std::endl;

   //! --- Set fastNLO verbosity level
   //say::SetGlobalVerbosity(say::DEBUG);

   // --- Set number of event
   double totwgt = vegasiterweight_.swgt;
   //double totwgt = runinfo_.nevents;
   double wgtfac = 1./runinfo_.itmax2;
   std::cout<<"[nnlo::term_fastnlo] VegasIterwgt:   "<<vegasiterweight_.swgt<<"\tniave="<<vegasiterweight_.iave<<"\tit="<<vegasiterweight_.it<<std::endl;
   std::cout<<"[nnlo::term_fastnlo] Runinfo itmax1: "<<runinfo_.itmax1<<"\titmax2="<<runinfo_.itmax2<<"\tnevents="<<runinfo_.nevents<<std::endl;
   std::cout<<"[nnlo::term_fastnlo] Total weight of table: "<<totwgt<<std::endl;
   std::cout<<"[nnlo::term_fastnlo] Weight factor: "<<wgtfac<<std::endl;

   if ( fnloUtils::ftable[id]==NULL ) {
      std::cout<<"[nnlo::term_fastnlo] WARNING! No event was filled into fastNLO table! Do not write empty file!"<<std::endl;
      return;
   }
   //std::string runid = runid_.srunid;
   fnloUtils::ftable[id]->MultiplyCoefficientsByConstant(totwgt * wgtfac);
   fnloUtils::ftable[id]->SetNumberOfEvents(totwgt); // *runinfo_.nshota

   // std::string filename = fnloUtils::ftablename[id] +"."+fnloUtils::GetOrderName( &fnloUtils::fNlNe )+"."+runid+".tab";
   // std::string filename = fnloUtils::ftablename[id] +"."+fnloUtils::GetOrderName( &fnloUtils::fNlNe )+".tab.gz";
   std::string filename = fnloUtils::ftablename[id]+".tab.gz";
   fnloUtils::ftable[id]->SetFilename(filename);

   // --- Write table
   if (fnloUtils::ftable[id]->GetIsWarmup())
      std::cout << "[nnlo::term_fastnlo] fastNLO warmup run finished ..." << std::endl;
   else
      std::cout << "[nnlo::term_fastnlo] fastNLO production run finished, writing table '" << fnloUtils::ftablename[id] << "'." << std::endl;

//   fnloUtils::ftable[id]->WriteTable(filename); // buggy. This needs a fix!
   fnloUtils::ftable[id]->WriteTable();

   return;
}



//-----------------------------------------------------------------------------
//
// --- --- --- fastNLO specific functions --- --- ---
//
//-----------------------------------------------------------------------------

// _____________________________________________________________________ //
fastNLO::GeneratorConstants  fnloUtils::GetNNLOJET_GenConsts() {
   //!< Get NNLOJET specific generator constants for this interface
   // The following is written to each contribution of a fastNLO table
   // Most of it should not be changed
   fastNLO::GeneratorConstants gc;
   // The code name normally should be set here
   gc.Name = "NNLOJET_rev_5088"; //!< Name and version of generator
   // The references to quote depend on the process; request to get them from NNLOJET
   // Alternatively update this default from the fastNLO steering card (to be implemented)
   gc.References.push_back("Example for single jet inclusive production."); //!< References for generator
   std::string pname = "Please adapt to selected process: " + nnlo::GetProcessName();
   gc.References.push_back(pname);
   gc.References.push_back("Please cite:"); //!< References for generator
   gc.References.push_back("Currie, J.; Glover, E.W.N. & Pires, J., Phys. Rev. Lett., 2017, 118, 072002 [arXiv:1611.01460]"); //!< References for generator
   // This should not be changed
   gc.UnitsOfCoefficients = 15; //!< NNLOJET provides cross sections in femtobarns fb
   return gc;
}



// _____________________________________________________________________ //
fastNLO::ProcessConstants fnloUtils::GetNNLOJET_ProcConstsPP() {
   //!< Get default process constants for pp processes
   // The following is filled mostly from NNLOJET and should not be changed
   std::vector<std::vector<int> > partLiCos = fnloUtils::sp.GetPartonCombinations();
   if ( partLiCos.empty() ) {
      std::cerr<<"[fnloUtils::GetNNLOJET_ProcConstsPP] Error. No parton combinations found. "
               << "Please read config file into subprocess-class: fnloUtils::sp. Aborting!"<<std::endl;
      exit(5);
   }

   fastNLO::ProcessConstants pc;
   pc.LeadingOrder      = nnlo::GetPowLO(); //!< Power in alpha_s of LO process
   pc.NPDF              = 2;                //!< No. of PDFs involved
   int nprocs = partLiCos.size();           //   These are derived and filled per process and order; so only one of LO, NLO, NNLO is active at any time
   pc.NSubProcessesLO   = nprocs;           //!< No. of LO   subprocesses
   pc.NSubProcessesNLO  = nprocs;           //!< No. of NLO  subprocesses
   pc.NSubProcessesNNLO = nprocs;           //!< No. of NNLO subprocesses
   pc.IPDFdef1          = 3;                //!< Flag 1 to define PDF linear combinations of partonic subprocesses
   pc.IPDFdef2          = 0;                //!< Flag 2 to define PDF linear combinations of partonic subprocesses
   pc.IPDFdef3LO        = nprocs;           //!< Flag 3 to define PDF LCs at   LO
   pc.IPDFdef3NLO       = nprocs;           //!< Flag 3 to define PDF LCs at  NLO
   pc.IPDFdef3NNLO      = nprocs;           //!< Flag 3 to define PDF LCs at NNLO
   pc.NPDFDim           = 2;                //!< Internal storage mode for PDF LCs, 2: full-matrix ==> we do not need 'asymmetric processes'
   //!< PDF Linear combinations in two different formats (used only if IPDFdef2==0) (Why two formats? TODO)
   pc.PDFCoeffLO        = VectorToPairs(partLiCos); //!< PDF Linear combinations in pairs
   pc.PDFCoeffNLO       = pc.PDFCoeffLO;
   pc.PDFCoeffNNLO      = pc.PDFCoeffLO;
   pc.PDFLiCoInLO       = partLiCos;                //!< PDF Linear combinations
   pc.PDFLiCoInNLO      = pc.PDFLiCoInLO;
   pc.PDFLiCoInNNLO     = pc.PDFLiCoInLO;
   pc.AsymmetricProcesses.clear();                  //!< Specify processes that need to be exchanged in half-matrix notation, when xmin>xmax (only if NPDFDim==1)
   pc.Name              = nnlo::GetProcessName();   //!< Name of process as given in NNLOJET runcard
   pc.References.push_back("PC: Please check and modify NNLOJET references according to selected process."); // references for processes
   return pc;
}



// _____________________________________________________________________ //
fastNLO::ProcessConstants fnloUtils::GetNNLOJET_ProcConstsDIS() {
   //!< Get default process constants for DIS processes
   // The following is filled mostly from NNLOJET and should not be changed
   std::vector<std::vector<int> > partLiCos = fnloUtils::sp.GetPartonCombinations();
   if ( partLiCos.empty() ) {
      std::cerr<<"[fnloUtils::GetNNLOJET_ProcConstsDIS] Error. No parton combinations found. "
               << "Please read config file into subprocess-class: fnloUtils::sp. Aborting!"<<std::endl;
      exit(5);
   }

   fastNLO::ProcessConstants pc;
   pc.LeadingOrder      = nnlo::GetPowLO(); //!< Power in alpha_s of LO process
   pc.NPDF              = 1;                //!< No. of PDFs involved
   int nprocs = partLiCos.size();           //   These are derived and filled per process and order; so only one of LO, NLO, NNLO is active at any time
   pc.NSubProcessesLO   = nprocs;           //!< No. of LO   subprocesses
   pc.NSubProcessesNLO  = nprocs;           //!< No. of NLO  subprocesses
   pc.NSubProcessesNNLO = nprocs;           //!< No. of NNLO subprocesses
   pc.IPDFdef1          = 2;                //!< Flag 1 to define PDF linear combinations of partonic subprocesses
   pc.IPDFdef2          = 0;                //!< Flag 2 to define PDF linear combinations of partonic subprocesses
   pc.IPDFdef3LO        = nprocs;           //!< Flag 3 to define PDF LCs at   LO
   pc.IPDFdef3NLO       = nprocs;           //!< Flag 3 to define PDF LCs at  NLO
   pc.IPDFdef3NNLO      = nprocs;           //!< Flag 3 to define PDF LCs at NNLO
   pc.NPDFDim           = 0;                //!< Internal storage mode for PDF LCs, 0: linear, only one PDF involved here
   //!< PDF Linear combinations in two different formats (used only if IPDFdef2==0) (Why two formats? TODO)
   pc.PDFCoeffLO        = VectorToPairs(partLiCos); //!< PDF Linear combinations in pairs
   pc.PDFCoeffNLO       = pc.PDFCoeffLO;
   pc.PDFCoeffNNLO      = pc.PDFCoeffLO;
   pc.PDFLiCoInLO       = partLiCos;                //!< PDF Linear combinations
   pc.PDFLiCoInNLO      = pc.PDFLiCoInLO;
   pc.PDFLiCoInNNLO     = pc.PDFLiCoInLO;
   pc.AsymmetricProcesses.clear();                  //!< Specify processes that need to be exchanged in half-matrix notation, when xmin>xmax (only if NPDFDim==1)
   pc.Name              = nnlo::GetProcessName();   //!< Name of process as given in NNLOJET runcard
   pc.References.push_back("PC: Please check and modify NNLOJET references according to selected process."); // references for processes
   return pc;
}



// _____________________________________________________________________ //
fastNLO::ScenarioConstants fnloUtils::GetNNLOJET_ScenConsts(const int& id) {
   //!< Set default scenario constants as appropriate with NNLOJET
   // The following is partially filled from NNLOJET
   // Some settings can/should be adapted from a steering file to modify what kind of table is created
   fastNLO::ScenarioConstants sc;

   //!> Scenario name set to default of NNLOJET process name
   //!> CHANGE in steering file to desired value; no white space allowed here!
   sc.ScenarioName = nnlo::GetProcessName();
   //!> Scenario description, empty by default
   //!> CHANGE in steering file to desired documentation lines
   sc.ScenarioDescription.clear();
   sc.ScenarioDescription.push_back("Please replace this by desired description of scenario");
   //!> For direct comparison to Rivet analyses this must be set to the units, in which the data were published
   //!> Usually LHC Rivet analyses are working in units of [pb] = 10^-12 [b]
   //!> CHANGE in steering file to required value
   sc.PublicationUnits = 12;
   //!> So far the interface to NNLOJET supports only 1-dim histograms
   //!> DO NOT CHANGE!
   sc.DifferentialDimension = 1;
   //!> fastNLO allows to define Labels (symbol and unit) for the measurement dimensions
   //!> (from outer to inner "loop"), e.g. "|y|" and "pT_[GeV]".
   //!> With NLOJet++ this helps to define the observables to be calculated in an automatized way!
   //!> Here we will set it via the gridname from the NNLOJET run card
   //!> CHANGE in steering file if desired
   const std::string& gridname = fnloUtils::fInits[id].gridname;
   size_t ipos = gridname.find_last_of("_");
   if (ipos < std::string::npos) ipos=ipos+1;
   std::string dimlabel = gridname.substr(ipos,std::string::npos);
   sc.DimensionLabels.clear();
   sc.DimensionLabels.push_back(dimlabel);
   //!> Specify for each dimension whether
   //!>  0 : the cross section is NOT differential,
   //!>      i.e. there are two bin borders (but NO division (normalization) by bin width);
   //!>  1 : the cross section is point-wise differential, i.e. only one point is given;
   //!>  2 : the cross section is bin-wise differential,   i.e. there are two bin borders and division by bin width
   //!> Since NNLOJET provides bin-wise differential distributions use option 2
   //!> DO NOT CHANGE!
   sc.DimensionIsDifferential.clear();
   sc.DimensionIsDifferential.push_back(2);
   //!> Calculate bin width from lower and upper bin boundaries; this is the fastNLO default
   //!> CHANGE in steering file to desired value
   sc.CalculateBinSize = true;
   //!> Possibility to provide additional normalization factor, e.g. of 2 for bins in |y|
   //!> The default, of course, is unity
   //!> CHANGE in steering file to desired value
   sc.BinSizeFactor = 1.;
   //!> If 'CalculateBinSize' is 'false', provide a vector with individual 'bin widths' for normalization.
   //!> This probably should only be done individually for each NNLOJET histo/fastNLO grid!
   //!> CHANGE in individual steering file unless only one histogram is filled by NNLOJET
   //!> sc.BinSize = ;
   //!> The binning is read from NNLOJET in warmup run and later reread via warmup files
   //!> DO NOT CHANGE HERE but in the NNLOJET run card
   //!> Scales and scale factors must be set in NNLOJET run card
   // TODO Could eventually be extracted from NNLOJET runcard (SCALES)
   //!> CHANGE ONLY FOR WARMUP RUN in steering file if desired
   sc.ScaleDescriptionScale1 = "scale1"; //< "<pT_1,2>_[GeV]" # This defines the scale to be used (Note: The 1st scale should always be in units of [GeV]!)
   sc.ScaleDescriptionScale2 = "scale2"; //< "pT_max_[GeV]"   # Specify 2nd scale name and unit (ONLY for flexible-scale tables)
   // TODO
   //!> The center-of-mass energy (in GeV) is read from NNLOJET
   //!> DO NOT CHANGE HERE but in the NNLOJET run card
   sc.CenterOfMassEnergy = inputpar_.roots;
   //!> IDs for PDF of 1st and (if any) 2nd colliding hadron (following PDG convention: proton 2212).
   //!> The colliding hadrons should be read from NNLOJET ... TODO
   //!> DO NOT CHANGE HERE but NNLOJET if possible
   sc.PDF1 = 2212;
   sc.PDF2 = 2212;
   //!> Filename of fastNLO output table; will be overwritten below per histogram id
   //!> DO NOT CHANGE in steering file unless only one histogram is filled by NNLOJET
   sc.OutputFilename = "fastNLO";
   //!> Number of decimal digits to store in output table (def.=8, which is normally fully sufficient).
   //!> DO NOT CHANGE in steering file unless really necessary
   sc.OutputPrecision = 8;
   //!> Assume that zlib is available and output compression can be used
   //!> CHANGE in steering file if compression not possible or not desired
   sc.OutputCompression = true;
   //!> Create fixed- or flexible-scale table
   //!>
   //!> Fixed-scale tables require to define one central scale plus a fixed set of mu_f scale factors (at least one, e.g. 1.0)
   //!> For each mu_f scale factor a copy of the interpolation grid is stored.
   //!> The mu_r dependence is evaluated analytically; the mu_f dependence can be evaluated analytically up to NLO using HOPPET.
   //!> Beyond that only the defined set of mu_f scale factors can be used
   //!>
   //!> Flexible-scale tables store separately weights up to the following scale dependences:
   //!> independent, log(mu_r), log(mu_f), log(mu_r)^2, log(mu_f)^2, log(mu_r)*log(mu_f)
   //!> It is possible to store two different settings for the central scale, but the first one must be in GeV.
   //!> It is possible to redefine the central scale separately for mu_r and mu_f to either scale1, scale2, or a function of the two like sqrt(mu_1^2 + mu_2^2)
   //!> Flexible-scale tables require more information from a generator, are larger in size, and somewhat slower than fixed-scale tables.
   //!> Possible with NNLOJET, but check carefully.
   //!> CHANGE in steering file if desired
   sc.FlexibleScaleTable = false;
   //!> Factorization scale variations must be set in accordance with NNLOJET SCALES settings.
   //!> Only needed for fixed-scale tables beyond LO.
   //!> List of scale factors must include factor '1.0'.
   //!> Scale factors will be ordered according to fastNLO convention: (1, min, ... , max). Default: {1.0, 0.5, 2.0}
   //!>
   sc.ScaleVariationFactors.clear();
   sc.ScaleVariationFactors.push_back(1.0);
   if ( ! nnlo::GetPowLO() ) {
     sc.ScaleVariationFactors.push_back(0.5);
     sc.ScaleVariationFactors.push_back(2.0);
   }
   //!> Specify if binning is read from fScenConst or from warmup; with NNLOJET binning from run card is written to warmup
   //!> DO NOT CHANGE
   sc.ReadBinningFromSteering = true;
   //!> Do not check once more the binning between warmup and steering; binning is taken from NNLOJET run card anyway
   //!> DO NOT CHANGE
   sc.IgnoreWarmupBinningCheck = true;
   //!> Apply reweighting of pdfs for an optimized interpolation, def.=true.
   //!> DO NOT CHANGE in steering file unless you really know what you do
   sc.ApplyPDFReweighting  = true;
   //!> Set limits for scale nodes to bin borders, if possible. Good if scale equals the binned observable.
   //!> CHANGE FOR WARMUP RUN in steering file if desired
   sc.CheckScaleLimitsAgainstBins = true;
   //!> # -------------------------------------------------------------------- #
   //!> #   Choose fastNLO interpolation kernels and distance measures
   //!> # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   //!> #   Currently implemented interpolation kernels
   //!> #     Catmull
   //!> #     Lagrange
   //!> #     OneNode
   //!> #     Linear
   //!> #
   //!> #   Currently implemented distance measures
   //!> #     linear
   //!> #     loglog025        eq. to (log(log(4*x)))
   //!> #     log10
   //!> #     sqrtlog10        eq. to sqrt(log_10(x))
   //!> # -------------------------------------------------------------------- #
   //!> CHANGE in steering file ONLY if required AFTER checking interpolation quality for an observable
   //!> If necessary probably should be changed individually for each NNLOJET histo/fastNLO grid using specific steering files
   //!>
   //!> x-interpolation (momentum fraction)
   sc.X_Kernel = "Lagrange";
   sc.X_DistanceMeasure = "sqrtlog10";
   sc.X_NNodes = 20;                   // APPLgrid setting is 30, but leads to very large tables
   sc.X_NNodeCounting = "NodesPerBin"; // Corresponds to fastNLO v14 & v21 standard
   if ( nnlo::IsDIS() ) {
      sc.X_Kernel = "Catmull";
      sc.X_DistanceMeasure = "log10";  // We like to have many nodes at LOW-x for DIS!
      sc.X_NNodes = 18;
   }
   //!> mu1-interpolation (scale)
   sc.Mu1_Kernel = "Lagrange";
   sc.Mu1_DistanceMeasure = "loglog025";
   sc.Mu1_NNodes = 6;
   //!> mu2-interpolation (scale); only for flexible-scale tables
   sc.Mu2_Kernel = "Lagrange";
   sc.Mu2_DistanceMeasure = "loglog025";
   sc.Mu2_NNodes = 6;

   //!> CHANGE in steering file in accordance with NNLOJET run card!
   sc.InclusiveJets = false;           // Use NNLOJET COMPOSITE histos to store multiple entries per event
                                       // However, in order to use multiple scale values per event, this flag must be set!

   return sc;
}


//-----------------------------------------------------------------------------
//
// --- --- --- fnloUtils --- --- --
//
//-----------------------------------------------------------------------------

// _____________________________________________________________________ //
void fnloUtils::SetNNLOJETDefaultBinning(fastNLO::ScenarioConstants& sc, const int& nbins, const double& lo) {
   //!< set binning as used by NNLOJET
   sc.SingleDifferentialBinning.resize(nbins+1);
   for (int i=0; i<=nbins; i++) sc.SingleDifferentialBinning[i] = (&lo)[i];
   std::cout<<"DefaultBinning: "<<lo;
   for (int i=1; i<=nbins; i++) std::cout<<"\t"<<sc.SingleDifferentialBinning[i];
   std::cout<<std::endl;
   return ;
   // double binwidth = (hi-lo)/nbins;
   // sc.SingleDifferentialBinning[0] = lo;
   // std::cout<<"DefaultBinning: "<<lo;
   // for (int i=1; i<=nbins; i++) {
   //    sc.SingleDifferentialBinning[i] = sc.SingleDifferentialBinning[i-1]+binwidth;
   //    std::cout<<"\t"<<sc.SingleDifferentialBinning[i];
   // }
   // std::cout<<std::endl;
}



// // _____________________________________________________________________ //
// std::string fnloUtils::GetProcessName() {
//    //!< get name of the process from NNLOJET
//    std::string process(process_.sproc);
//    process.erase(std::remove(process.begin(), process.end(), ' '),process.end());
//    return process;
// }



// _____________________________________________________________________ //
std::string fnloUtils::GetContribName(int nloops, int ne) {
   //!<
   //!< Get name of contributions
   //!<

 /*
 * LO = #looops=0  #extra-emissions=0
 * V = #looops=1  #extra-emissions=0
 * R = #looops=0  #extra-emissions=1
 * VV = #looops=2  #extra-emissions=0
 * RV = #looops=1  #extra-emissions=1
 * RR = #looops=0  #extra-emissions=2
 */

   if ( nloops==-1 && ne==-1 ) {
      std::cout<<"Works only for having a CURRENT event !!!"<<std::endl;
      nloops = ndimcurrent_.nv;
      static const int& nord   = nnlo::GetOrder();
      // int njet     = inppar_.njets;
      // if ( nnlo::IsDIS() ) njet -= 1;
      ne = nord - nloops;
   }
   if ( nloops == 0 && ne == 0) return "LO";
   else if ( nloops == 1 && ne == 0) return "V";
   else if ( nloops == 0 && ne == 1) return "R";
   else if ( nloops == 2 && ne == 0) return "VV";
   else if ( nloops == 1 && ne == 1) return "RV";
   else if ( nloops == 0 && ne == 2) return "RR";
   else return "Contrib Name NOT IDENTIFIED: "+std::to_string(nnlo::GetOrder());

}



// // _____________________________________________________________________ //
// int fnloUtils::GetOrder() {
//    static const int& nord = ndimcurrent_.norder ;
//    static const int& nv   = ndimcurrent_.nv ;
//    return nord + nv;
// }



// _____________________________________________________________________ //
std::string fnloUtils::GetOrderName( std::set<std::pair<int,int> >* nlne ) {
   //!< get order of calculation as 'string'

   if ( nlne == NULL) {
      int nord   = nnlo::GetOrder();
      //static const int& nord   = ndimcurrent_.norder;
      if ( nord == 0 ) return "LO";
      else if ( nord == 1 ) return "NLO";
      else if ( nord == 2 ) return "NNLO";
   }
   else if ( fnloUtils::fIDs.size()==1 ) {
      // only one channel was calculated
      // take:  'chZYZ'
      char buf[5];
      sprintf(buf, "%03d", *fnloUtils::fIDs.begin() );
      return "ch"+std::string(buf);
   }
   else {
      std::set<std::string> names;
      for ( auto i : *nlne ) names.insert(fnloUtils::GetContribName(i.first,i.second));
      if ( names.count("R") && names.count("V") ) {
         names.erase("R");
         names.erase("V");
         names.insert("NLO");
      }
      if ( names.count("RR") && names.count("VV") && names.count("RV") ) {
         names.erase("RR");
         names.erase("RV");
         names.erase("VV");
         names.insert("NNLO");
      }
      std::string ret;
      for ( auto i : names )
         if ( ret.empty() ) ret = i;
         else ret += "+"+i;
      return ret;
   }
   return ""; //?
}


// _____________________________________________________________________ //
int fnloUtils::GetMaxscl() {
   return maxscl_hook; // nnlo_common.
}

// _____________________________________________________________________ //
int fnloUtils::GetSeed() {
   //!< get seed value

   // new
   return iseed_hook; // nnlo_common.

   // old
   /*
   std::vector<std::string> tok;
   char clog[101];
   strcpy(clog,log_.slogname);
   char *p = strtok(clog, ".");
   while (p) {
      tok.push_back(p);
      p = strtok(NULL, ".");
   }
   if ( tok.size()<=2 ) return -1;
   std::string sN = tok[tok.size()-2];
   sN.erase(0,1);
   int ret = atoi( sN.c_str());
   return ret;
   */
}



// _____________________________________________________________________ //
std::string fnloUtils::GetConfigDir() {
   //!< Get directory where config files are stored
   // CONFIGDIR is set by NNLOJET at 'make' time! Later changes per environment variable not possible!
   std::string configdir(getenv("CONFIGDIR")); // Read from NNLOJET setting
   std::cout << "[fnloUtils::GetConfigDir()] CONFIGDIR of NNLOJET found to be: " << configdir << std::endl;
   if ( configdir == "" ) {
      std::cout << "[fnloUtils::GetConfigDir()] WARNING! CONFIGDIR of NNLOJET not set, "
                << "using default $PWD/process instead." << std::endl;
      configdir = std::string(getenv("PWD"))+"/process";
   }
   // --- test config directory
   struct stat myStat;
   std::cout << "[fnloUtils::GetConfigDir()] Status of CONFIGDIR found to be: " << stat(configdir.c_str(), &myStat) << std::endl;
   if ( !((stat(configdir.c_str(), &myStat) == 0) && (((myStat.st_mode) & S_IFMT) == S_IFDIR)) ) {
      std::cout << "[fnloUtils::GetConfigDir()] WARNING! Could not find the config directory " <<
         configdir << std::endl;
      // try default in case CONFIGDIR was set differently above
      configdir = std::string(getenv("PWD"))+"/process";
      std::cout << "[fnloUtils::GetConfigDir()] Trying default setting for CONFIGDIR of NNLOJET: " << configdir << std::endl;
      if ( !((stat(configdir.c_str(), &myStat) == 0) && (((myStat.st_mode) & S_IFMT) == S_IFDIR)) ) {
         std::cout << "[fnloUtils::GetConfigDir()] WARNING! Could not find the config directory " <<
            configdir << std::endl;
         // another and next to last guess (Grid site)
         configdir = "/cvmfs/etp.kit.edu/nnlo/src/driver/process";
         std::cout << "[fnloUtils::GetConfigDir()] Trying grid site setting for CONFIGDIR: " << configdir << std::endl;
         if ( !((stat(configdir.c_str(), &myStat) == 0) && (((myStat.st_mode) & S_IFMT) == S_IFDIR)) ) {
            std::cout << "[fnloUtils::GetConfigDir()] WARNING! Could not find the config directory " <<
               configdir << std::endl;
            // really last guess (local user install); procedure to be rethought ...
            configdir = std::string(getenv("HOME"))+"/local/src/NNLOJET-rev3678/driver/process";
            std::cout << "[fnloUtils::GetConfigDir()] Trying local install setting for CONFIGDIR: " << configdir << std::endl;
            if ( !((stat(configdir.c_str(), &myStat) == 0) && (((myStat.st_mode) & S_IFMT) == S_IFDIR)) ) {
               std::cerr << "[fnloUtils::GetConfigDir()] ERROR! Really could not find a config directory " <<
                  configdir << std::endl;
               exit(2);
            }
         }
      }
   }
   std::cout<<"[fnloUtils::GetConfigDir()] Using config directory: "<<configdir<<std::endl;
   return configdir;
}



// _____________________________________________________________________ //
std::string fnloUtils::GetConfigFile() {
   //!< get name of the process
   std::string pname = nnlo::GetProcessName();
   std::cout << "[fnloUtils::GetConfigFile()] The process name is reported to be: " << pname << std::endl;
   if ( pname == "1jet" || pname == "2jet" ) {
     pname = pname.substr(1,std::string::npos);
     std::cout << "[fnloUtils::GetConfigFile()] For subprocess configuration shortening process name to: " << pname << std::endl;
   }
   std::string fname = pname + "-" + fnloUtils::GetOrderName() + ".config";
   fname = fnloUtils::GetConfigDir() + "/" + pname + "/" + fname;
   std::cout << "[fnloUtils::GetConfigFile()] Using config file: " << fname << std::endl;
   return fname;
}



// _____________________________________________________________________ //
std::vector<std::string> fnloUtils::GetConfigFiles() {
   //!< get name of the process
   std::string pname = nnlo::GetProcessName();
   std::cout << "[fnloUtils::GetConfigFiles()] The process name is reported to be: " << pname << std::endl;
   if ( pname == "1jet" || pname == "2jet" ) {
     pname = pname.substr(1,std::string::npos);
     std::cout << "[fnloUtils::GetConfigFiles()]  For subprocess configuration shortening process name to: " << pname << std::endl;
   }
   std::vector<std::string> fnames;
   std::string pref = fnloUtils::GetConfigDir()+"/"+pname+"/"+pname;
   if ( nnlo::GetOrder() == 0 ) {
      fnames.push_back( pref+"-LO.config");
   }
   else if ( nnlo::GetOrder() == 1 ){
      fnames.push_back( pref+"-R.config");
      fnames.push_back( pref+"-V.config");
   }
   else if ( nnlo::GetOrder() == 2 ){
      fnames.push_back( pref+"-RR.config");
      fnames.push_back( pref+"-RV.config");
      fnames.push_back( pref+"-VV.config");
   }
   std::cout<<"[fnloUtils::GetConfigFiles()]  Found order "<<nnlo::GetOrder()<<". Using config files: "<<std::endl;
   for ( auto f : fnames ) std::cout<<"\t"<<f<<std::endl;
   return fnames;
}



// _____________________________________________________________________ //
void fnloUtils::RememberContribution() {
   //!< fill the variable fNlNe
   //!< and remember the calculated contributions

   static const int& iproc = currentprocess_.iproc;
   int nloops = channel_.iloops[iproc-1];
   int nreal  = ndimcurrent_.norder;

   // static const int& nloops = ndimcurrent_.nv;
   // int nord                 = nnlo::GetOrder();//ndimcurrent_.norder;
   // // int njet                 = inppar_.njets;
   // // if ( nnlo::IsDIS() ) njet -= 1;
   // // int ne = nord - nloops;
   // //std::cout<<"nl="<<nloops<<"\tne="<<ne<<"\t\tnord="<<nord<<"\tIsDIS="<<nnlo::IsDIS()<<"\tnjet="<<njet<<std::endl;
   fnloUtils::fNlNe.insert({ nloops,nreal });
   fnloUtils::fIDs.insert(iproc);

}



// _____________________________________________________________________ //
double fnloUtils::CalculateReferenceWeight(double wgt, double x1, double x2) {
   //!< calculate the reference weight
   //!< and compare it later with the ref-weight from NNLOJET
   //!<
   //!< Set value: _myfillwgt
   //!< Need access to evolvepdf_()
   //!< needs either strong_.as or alphaspdf_()
   //!<

   double PDF = 0;
   double muf2  = muf2_hook(1);
   const std::vector<std::vector<int > >&  pc =  fnloUtils::sp.GetInitialValues();
   if ( nnlo::IsDIS() ) {
      double xfx[13] = {0};
      evolvepdf_(x1,sqrt(muf2),xfx);
      for ( unsigned int ip = 0 ; ip < pc[currentprocess_.iproc].size() ; ip+=2 ) {
            PDF += xfx[pc[currentprocess_.iproc][ip]+6];
      }
   }
   else {
      double xfx1[13] = {0};
      double xfx2[13] = {0};
      evolvepdf_(x1,sqrt(muf2),xfx1);
      evolvepdf_(x2,sqrt(muf2),xfx2);
      for ( unsigned int ip = 0 ; ip < pc[currentprocess_.iproc].size() ; ip+=2 ) {
         PDF += xfx1[pc[currentprocess_.iproc][ip]+6]*xfx2[pc[currentprocess_.iproc][ip+1]+6];
      }
   }

   static const double& as = strong_.as; //alphaspdf_(sqrt(mur2));
   _myfillwgt = wgt * PDF * pow( as/(2*M_PI) , nnlo::GetNPow() ) ;
   return _myfillwgt;
}



// _____________________________________________________________________ //
std::vector<double> fnloUtils::SclIndepWgt(const double* wt, double muf, double mur) {
   //! Calculate scale independent weights for fastNLO

   // Do not use debug mode for mass production!
   static const bool Debug = false;
   static bool First = true;
   if ( First && Debug ) {
      First = false;
      std::cout << "[fnloUtils::SclIndepWgt]: Running in DEBUG mode. Do not use for mass production!" << std::endl;
   }

   std::vector<double> ret{0,0,0,0,0,0};  // return value
   std::vector<double> sret{0,0,0,0,0,0}; // simple sum
   std::vector<double> oret{0,0,0,0,0,0}; // ordered sum
   std::vector<double> kret{0,0,0,0,0,0}; // Kahan sum

   if ( nnlo::GetOrder()==0 ) {
      //! LO does not have scale dependent terms
      sret[0] = wt[0];
      oret[0] = wt[0];
      kret[0] = wt[0];
      ret[0]  = kret[0];
      return ret;
   }

   double wamax = DBL_MAX;
   const double EPS   = 1.e-3;  // aim for permille precision
   const double SMALL = 1.e-12; // relative weight limit to consider some deviations > EPS negligible
   static const unsigned int nSclVNNLO = 6;
   const unsigned int nSclV = nSclVNNLO;
   // Simplify in case of NLO, to be checked!!
   // unsigned int nSclV = nnlo::GetOrder()==1  ? 3 : 6; // only NLO

   if ( nSclV==3 ) {
      std::cerr << "[fnloUtils::SclIndepWgt]: Do not use yet NLO with nSclV = " << nSclV << "! Aborted!" << std::endl;
      exit(1);
      // -- not working, to be checked!
      // static const double SclInvNLO[3][3] = {{3, -2, 0},{0, 1, -1},{-1, 0, 1}}; // i+1
      // for ( unsigned int j = 0 ; j<nSclV ; j++ ){
      //    for ( unsigned int i = 0 ; i<nSclV ; i++ ){
      //       ret[j] += wt[i+1]*SclInvNLO[j][i];
      //    }
      //    //if ( fabs(ret[j]/wt[1]) < EPS ) ret[j] = 0;
      // }
      // -- not working, to be checked!
      // const double SclInvNLO[3][3] = {{5.5,-2,-2.5},{-1.5,1,0.5},{-0.5,0,0.5}}; // i+3
      // for ( unsigned int j = 0 ; j<nSclV ; j++ ){
      //         for ( unsigned int i = 0 ; i<nSclV ; i++ ){
      //            ret[j] += wt[i+3]*SclInvNLO[j][i];
      //         }
      // }
   } else {
      // This decomposition matrix is based on i.a. fixed scale choices of
      // exp^1, exp^(3/2), exp^(5/2);
      static const double SclInvNNLO[nSclVNNLO][nSclVNNLO] = {
         {    13.,  4.,    -9.,    -9.,     1., 1.},
         {-14./3., -2.,     2., 11./2.,     0., -5./6.},
         {-14./3., -2., 11./2.,     2., -5./6., 0.},
         {  1./3.,  0.,     0., -1./2.,     0., 1./6.},
         {  1./3.,  0., -1./2.,     0.,  1./6., 0.},
         {     1.,  1.,    -1.,    -1.,     0., 0.} };
      double sum[nSclVNNLO] = {0., 0., 0., 0., 0., 0.};
      int nRet = ( nnlo::GetOrder()==1 ) ? 4 : nSclV;

      // static unsigned int nprnt = 0;
      // static unsigned int npmod = 1;
      // if ( nprnt < 10 && nprnt % npmod == 0 ) {
      //    nprnt++;
      //    for ( int i = 0 ; i<=nRet ; i++ ){
      //       std::cout << std::setprecision(16) << "AAA i = " << i << ", wt[i] = " << wt[i] << std::endl;
      //    }
      // } else if ( nprnt == 10 ) {
      //    nprnt = 0;
      //    npmod *= 10;
      // }

      wamax = std::abs(wt[0]);
      for ( int j = 0 ; j<nRet ; j++ ){
         wamax = std::max(std::abs(wt[j+1]),wamax);
         double wasum = 0.;
         for ( unsigned int i = 0 ; i<nSclV ; i++ ){ // DB
            //         for ( unsigned int i = 0 ; i<nRet ; i++ ){ // KR TEST
            sum[i]   = wt[i+1]*SclInvNNLO[j][i];
            //            sret[j] += sum[i];
            wasum   += std::abs(sum[i]);
         }
         //         oret[j]  = fnloUtils::OrderedSum(sum,nSclVNNLO);
         kret[j]  = fnloUtils::KahanSum(sum,nSclVNNLO);
         // std::cout << "RD1 j = " << j << ", normal sum[j]  = " << sret[j] << std::endl;
         // std::cout << "RD2 j = " << j << ", ordered sum[j] = " << oret[j] << ", o/s = " << oret[j]/sret[j] << std::endl;
         // std::cout << "RD3 j = " << j << ", Kahan sum[j]   = " << kret[j] << ", k/s = " << kret[j]/sret[j] <<std::endl;
         // Numerically Kahan summing is the most precise
         ret[j] = kret[j];

         if ( Debug ) {
            // Set small weights to zero? Not necessary. Does not occur anyway for DBL_MIN ...
            static unsigned int nzwgt = 0;
            static unsigned int nzmss = 0;
            static unsigned int nzmod = 1;
            if ( ret[j] != 0. && std::abs(ret[j])/wasum < DBL_MIN ) {
               nzwgt++;
               if ( nzmss < 10 && nzwgt % nzmod == 0 ) {
                  nzmss++;
                  std::cout << "[fnloUtils::SclIndepWgt]: " << nzwgt << "th occurrence of tiny scale weight = " << ret[j] << std::endl;
               } else if ( nzmss == 10 ) {
                  nzmss = 0;
                  nzmod *= 10;
               }
            }
         }
      }

      // if ( nnlo::GetOrder()==1 ) {
      //    // ret[3]=0;
      //    ret[4]=0;// must be exact zero for fastNLO
      //    ret[5]=0;// must be exact zero for fastNLO
      // }

   }

   // Try to reconstruct original weight wt[0] as cross check
   if ( Debug ) {
      double r = log(mur*mur);
      double f = log(muf*muf);
      double sum[6] = { ret[0], r*ret[1], f*ret[2], r*r*ret[3], f*f*ret[4], r*f*ret[5] };
      // double sMyWgt0 = 0.;
      // for ( unsigned int i=0; i<nSclVNNLO; i++) {
      //    sMyWgt0 += sum[i];
      // }
      // double oMyWgt0 = OrderedSum(sum,nSclVNNLO);
      double kMyWgt0 = KahanSum(sum,nSclVNNLO);
      // Numerically Kahan summing is the most precise
      double MyWgt0  = kMyWgt0;

      //double MyWgtNLO   = ret[0] + r*ret[1] + f*ret[2];
      //double MyRefWgt = ret[0] + 2*r*ret[1] + 2*f*ret[2] + 4*r*r*ret[3] + 4*f*f*ret[4] + 4*r*f*ret[5];

      // Count misreconstructed weights for first scale
      static unsigned int nmwgt = 0;
      static unsigned int nmmss = 0;
      static unsigned int nmmod = 1;
      if  ( std::abs(MyWgt0/wt[0]-1) > EPS || !std::isfinite(MyWgt0)) {
         if ( std::abs(MyWgt0)/wamax > SMALL || std::abs(wt[0])/wamax > SMALL ) {
            nmwgt++;
            if ( nmmss < 10 && nmwgt % nmmod == 0 ) {
               nmmss++;
               std::cout << "[fnloUtils::SclIndepWgt]: " << nmwgt << "th occurrence of large misreconstructed weight!" << std::endl;
               std::cout<<"[fnloUtils::SclIndepWgt]: LARGE discrepant weight! Original weight = " << wt[0] << ", reconstructed weight = " << MyWgt0 << ", deviation (per mille) = " << 1000*(std::abs(MyWgt0/wt[0])-1) << ", weight max = " << wamax << std::endl;
               std::cout<<"[fnloUtils::SclIndepWgt]: Original weight/max = " << std::abs(wt[0])/wamax << ", reconstructed weight/max = " << std::abs(MyWgt0)/wamax << std::endl;
            } else if ( nmmss == 10 ) {
               nmmss = 0;
               nmmod *= 10;
            }
         }
         std::cerr<<"[fnloUtils::SclIndepWgt]. Warning!"<<std::endl;
         std::cerr<<"[fnloUtils::SclIndepWgt]. Discrepant weight! Original weight = " << wt[0] << ", reconstructed weight = " << MyWgt0 << ", weight max = " << wamax << std::endl;
         std::cerr<<"[fnloUtils::SclIndepWgt]. Original weight/max = " << std::abs(wt[0])/wamax << ", reconstructed weight/max = " << std::abs(MyWgt0)/wamax << std::endl;
         std::cout<<"  muf = " << muf << ", mur = " << mur << std::endl;
         std::cout<<"  wt [] =";
         for ( unsigned int i = 0 ; i<7 ; i++ )  std::cout<<"\t"<<wt[i];
         std::cout<<"\t\t"<<std::endl;

         std::cout<<"  ln(mur^2) []";
         for ( unsigned int i = 0 ; i<7 ; i++ )  std::cout<<"\t"<<log(mur2_hook(i+1));
         std::cout<<"\t\t"<<std::endl;

         std::cout<<"  ln(muf^2) []";
         for ( unsigned int i = 0 ; i<7 ; i++ )  std::cout<<"\t"<<log(muf2_hook(i+1));
         std::cout<<"\t\t"<<std::endl;

         std::cout<<"  ret[]";
         for ( unsigned int j = 0 ; j<6 ; j++ )std::cout<<"\t"<<ret[j];
         std::cout<<std::endl;
         std::cout<<" oret[]";
         for ( unsigned int j = 0 ; j<6 ; j++ )std::cout<<"\t"<<oret[j];
         std::cout<<std::endl;
         std::cout<<" kret[]";
         for ( unsigned int j = 0 ; j<6 ; j++ )std::cout<<"\t"<<kret[j];
         std::cout<<std::endl;

         //         exit(3);
      }
   }

   return ret;
}


// Kahan summing to improve numerical precision
double fnloUtils::KahanSum( double in[], unsigned int insize ) {
   double ksum = in[0];
   double c = 0.;
   for ( unsigned int i = 1; i<insize ; i++ ){
      double y = in[i] - c;
      double t = ksum + y;
      c        = (t  - ksum)  - y;
      ksum     = t;
   }
   return ksum;
}


// Ordered summing (in descending order here) to improve numerical precision
double fnloUtils::OrderedSum( double in[], unsigned int insize ) {
   double osum = 0.;
   std::vector<double> vin;
   for ( unsigned int i = 0 ; i<insize ; i++ ){
      vin.push_back(in[i]);
   }
   auto func=[](double a, double b) { return std::abs(a) > std::abs(b); };
   std::sort(std::begin(vin), std::end(vin), func);
   for ( unsigned int i = 0 ; i<insize ; i++ ){
      osum += vin[i];
   }
   return osum;
}



// _____________________________________________________________________ //
void fnloUtils::InitFastNLO( const int& id, const double* wt ) {
   // --- Debug printout of NNLOJET input
   // static const int& iproc = currentprocess_.iproc;
   // int nloops = channel_.iloops[iproc-1]; /// loop order
   // int ipars  = channel_.ipars[iproc-1];  /// number of additional particles + 1
   // std::cout<<"        process_.iproc  "<<currentprocess_.iproc<< std::endl;
   // std::cout<<"        nloops          "<<nloops<< std::endl;
   // std::cout<<"        ipars           "<<ipars<< std::endl;
   // std::cout<<"        ndim_.norder    "<<ndimcurrent_.norder<< std::endl;
   // std::cout<<"        inppar_.njets   "<<inppar_.njets<< std::endl;
   // std::cout<<"        order_.ieorder  "<<order_.ieorder<< std::endl;
   // std::cout<<"        GetNPow         "<<nnlo::GetNPow()<< std::endl;
   // std::cout<<"        GetOrder()      "<<nnlo::GetOrder()<< std::endl;
   // std::cout<<"        GetMaxscl()     "<<fnloUtils::GetMaxscl()<<", maxscl = "<<maxscl<< std::endl;
   // for (int is = 1; is<maxscl+1; is++) {
   //   std::cout<<"        is = " << is << ", mur(is) = " << sqrt(mur2_hook(is)) << ", muf(is) = " << sqrt(mur2_hook(is)) << std::endl;
   //   //    std::cout<<"        is = " << is << ", xmur(is) = " << scales_.rscale[is] << ", xmuf(is) = " << scales_.facscale[is] << std::endl;
   // }
   // std::cout<< std::endl;

   // --- Welcome
   std::cout << "[fnloUtils::InitFastNLO] fastNLO initialisation for NNLOJET histo/fastNLO grid with id = " << id << std::endl;
   if ( fnloUtils::ftable[id] != NULL )  return;
   _myfillwgt = 0;
   std::cout << "[fnloUtils::InitFastNLO] NNLOJET provides process: " << nnlo::GetProcessName() << std::endl;

   // --- Get settings of histogram [id] from NNLOJET
   const std::string& gridname = fnloUtils::fInits[id].gridname;
   const int& nbins = fnloUtils::fInits[id].nbins;
   const double& lo = *(fnloUtils::fInits[id].lo);
   //const double& hi = *(fnloUtils::fInits[id].hi);

   // --- Read config file with  parton combinations
   if ( fnloUtils::sp.GetConfigFile().empty() )
     fnloUtils::sp.ReadConfigFiles(fnloUtils::GetConfigFiles());

   // --- Collect info for naming
   std::string wrmfile;
   std::string steerfile;
   std::string gsteerfile;
   std::string tablename;
   std::string runid = runid_.srunid;
   std::string pname = nnlo::GetProcessName();
   std::string jname = jobname_.sjobname;
   jname.erase(remove(jname.begin(), jname.end(), ' '),jname.end());
   std::string fname = outfile_.fname;
   //   fname.erase(remove(fname.begin(), fname.end(), ' '),fname.end());
   size_t ipos = fname.find_first_of(' ');
   fname.erase(ipos, std::string::npos);
   int seed = fnloUtils::GetSeed();
   std::cout << "[fnloUtils::InitFastNLO] Gridname: " << gridname << std::endl;
   std::cout << "[fnloUtils::InitFastNLO] Integration input: " << fname << std::endl;
   std::cout << "[fnloUtils::InitFastNLO] Process name: " << pname << ", job name: " << jname << std::endl;
   // If unitphase, i.e. grid warmup, then name wrmfile individually for each subprocess for later check & merge
   if ( getunitphase_() ) {
     wrmfile  = pname+"."+jname+"."+gridname;
     wrmfile += ".wrm";
     std::cout << "[fnloUtils::InitFastNLO] Warmup is run for subprocess: " << jname << std::endl;
     std::cout << "[fnloUtils::InitFastNLO] Warmup filename: " << wrmfile << std::endl;
   }
   // If NOT unitphase, i.e. production, then expect identical wrmfile for all subprocesses
   else {
     wrmfile  = pname+"."+gridname;
     wrmfile += ".wrm";
     std::cout << "[fnloUtils::InitFastNLO] NOTE: The warmup file for production is expected to be identical for all subprocesses!" << std::endl;
     std::cout << "[fnloUtils::InitFastNLO]       The expected warmup filename therefore does not contain any subprocess label (job name) anymore. Please check!" << std::endl;
     std::cout << "[fnloUtils::InitFastNLO] Expected warmup filename: " << wrmfile << std::endl;
   }
   ipos = gridname.find_first_of("_");
   std::string scenname = gridname.substr(0,ipos);
   // Steerfile with settings specific to a NNLOJET histo/fastNLO grid
   steerfile = pname+"."+gridname;
   steerfile += ".str";
   // Global steerfile with settings valid for all NNLOJET histos/fastNLO grids
   // Values already set via specific steerfile are KEPT --> WARNINGS
   gsteerfile = pname+"."+scenname;
   gsteerfile += ".str";
   tablename = pname+"."+jname+"."+runid+"."+gridname+".s"+std::to_string(seed);
   std::cout << "[fnloUtils::InitFastNLO] Steering filename: " << steerfile << std::endl;
   std::cout << "[fnloUtils::InitFastNLO] Global steering filename: " << gsteerfile << std::endl;
   std::cout << "[fnloUtils::InitFastNLO] Tablename: " << tablename << std::endl;

   // --- Set NNLOJET specific default 'GeneratorConstants'
   fastNLO::GeneratorConstants gc = fnloUtils::GetNNLOJET_GenConsts();

   // --- Set process specific default 'ProcessConstants'
   // DIS or pp?
   fastNLO::ProcessConstants pc = nnlo::IsDIS() ? fnloUtils::GetNNLOJET_ProcConstsDIS() : fnloUtils::GetNNLOJET_ProcConstsPP();

   // --- Set scenario specific default 'ScenarioConstants'
   fastNLO::ScenarioConstants sc = fnloUtils::GetNNLOJET_ScenConsts(id);
   // --- Set scenario binning from NNLOJET info
   fnloUtils::SetNNLOJETDefaultBinning(sc, nbins, lo);

   // --- If existing read fastNLO steering file with settings specific for NNLOJET histo/fastNLO grid [id]
   bool fexist = !access(steerfile.c_str(), R_OK);
   if ( fexist) READ_NS(steerfile,wrmfile);
   //   read_steer::print(wrmfile);

   // --- The following ugly hacks are now replaced via steering files ==> to be removed
   if ( gridname.find("iff") != std::string::npos ) {
      std::cerr << "[fnloUtils::InitFastNLO] DEPRECATED! Please change fastNLO interpolation using steering file mechanism!" << std::endl;
      std::cerr << "[fnloUtils::InitFastNLO] Do not rely on presence of \"iff\" string in gridname anymore." << std::endl;
      std::cerr << "[fnloUtils::InitFastNLO] Special settings are ignored." << std::endl;
      //      sc.X_NNodes = 50;
      //      sc.X_DistanceMeasure = "sqrtlog10"; // we like to have increasingly many nodes at high-x
   }
   if ( nnlo::IsDIS() && gridname.find("q2") == std::string::npos ) {
      std::cerr << "[fnloUtils::InitFastNLO] DEPRECATED! Please change fastNLO interpolation using steering file mechanism!" << std::endl;
      std::cerr << "[fnloUtils::InitFastNLO] Do not rely on presence of \"q2\" string in gridname anymore." << std::endl;
      std::cerr << "[fnloUtils::InitFastNLO] Special settings are ignored." << std::endl;
      //      std::cout<<"[fnloUtils::InitFastNLO] IsDIS, but not differential as function of q2, using more q2 nodes, less x-nodes"<<std::endl;
      //      sc.Mu1_NNodes = 14; // more q2 nodes
      //      sc.Mu2_NNodes = 6;
      //      if ( gridname.find("iff") != std::string::npos ) sc.X_NNodes = 42;
   }
   // DEPRECATED: Fix hack to select between flexible- and fixed-scale table
   // std::cout<<"[fnloUtils::InitFastNLO] NNLOJET scale choices: muf2(1) = " << muf2_hook(1) << "\tmur2(1) = " << mur2_hook(1) << std::endl;
   // std::cout<<"[fnloUtils::InitFastNLO] NNLOJET scale choices: muf2(2) = " << muf2_hook(2) << "\tmur2(2) = " << mur2_hook(2) << std::endl;
   // if ( muf2_hook(2) == mur2_hook(2) && fabs(sqrt(mur2_hook(2))/exp(1)-1)<1.e-6) {
   //    std::cout<<"[fnloUtils::InitFastNLO] Flexible-scale table chosen."<< std::endl;
   //    sc.FlexibleScaleTable = true;
   // } else if ( muf2_hook(1) == mur2_hook(1) ) {
   //    std::cout<<"[fnloUtils::InitFastNLO] Fixed-scale table chosen."<< std::endl;
   //    sc.FlexibleScaleTable = false;
   // } else {
   //    std::cout<<"[fnloUtils::InitFastNLO] Unrecognized scale settings. Please choose either a fixed-scale or a flexible-scale table."<< std::endl;
   //    std::cout<<" Exiting."<< std::endl;
   //    exit(4);
   // }

   // DEPRECATED
   // fnloUtils::fIsInclJet[id] = (gridname.find("ji") != std::string::npos);
   // fnloUtils::fIs820GeV[id] = (gridname.find("820") != std::string::npos);
   // fnloUtils::fIs137[id]    = (gridname.find("137") != std::string::npos);

   // --- Initialise fastNLOCreate, use table object from now on
   sc.OutputFilename = tablename;
   fnloUtils::ftablename[id] = tablename;
   fnloUtils::ftable[id] = new fastNLOCreate(gc,pc,sc,wrmfile,gsteerfile);

   // --- Print out all steering settings so far for this table id/wrmfile
   read_steer::print(wrmfile);

   // --- Test consistency of fastNLO warmup and UNIT_PHASE setting in NNLOJET
   if ( fnloUtils::ftable[id]->GetIsWarmup() && !getunitphase_() ) {
      std::cerr<<"[fnloUtils::InitFastNLO] ERROR! No warmup file found, hence this cannot be a production run, but the job is NOT running with UNIT_PHASE. Aborted!"<< std::endl;
      std::cerr<<"[fnloUtils::InitFastNLO] Please recall that warmup files are produced and named per subprocess."<< std::endl;
      std::cerr<<"[fnloUtils::InitFastNLO] Hence they have to be merged and renamed to: "<< wrmfile << std::endl;
      exit(5);
   } else if ( !fnloUtils::ftable[id]->GetIsWarmup() && getunitphase_() ) {
      std::cerr<<"[fnloUtils::InitFastNLO] ERROR! Warmup file found, hence this is a production run, but the job is running WITH UNIT_PHASE. Aborted!"<< std::endl;
      exit(6);
   }

   // --- Further consistency tests
   bool isflex = fnloUtils::ftable[id]->GetIsFlexibleScale();
   bool isincljets = fnloUtils::ftable[id]->GetIsInclusiveJets();
   if ( isflex ) {
     std::cout << "[fnloUtils::InitFastNLO] The flexible-scale table format has been chosen." << std::endl;
   } else {
     std::cout << "[fnloUtils::InitFastNLO] The fixed-scale table format has been chosen." << std::endl;
   }
   std::cout << "[fnloUtils::InitFastNLO] Power of alpha_s(): " << nnlo::GetNPow() << ", order of calculation: " << nnlo::GetOrder() << std::endl;
   fnloUtils::ftable[id]->SetOrderOfAlphasOfCalculation(nnlo::GetNPow());
   fnloUtils::fIsInclJet[id] = isincljets;
   fnloUtils::ftable[id]->SetCacheSize(30);
   // For flexible-scale tables; TODO reactivate some checks!
   if ( isflex ) {
     // double q2 = muf2_hook(1);
     // double p2 = mur2_hook(1);
     // std::cout<<"mur: ";
     // for ( int is = 0 ; is<maxscl ; is++ ) std::cout<<"\t"<<sqrt(mur2_hook(1+is)) ;
     // std::cout<<std::endl;
     // std::cout<<"muf: ";
     // for ( int is = 0 ; is<maxscl ; is++ ) std::cout<<"\t"<<sqrt(muf2_hook(1+is)) ;
     // std::cout<<std::endl;

     // double muf2[7];
     // double mur2[7];
     // for ( int is = 0 ; is<maxscl ; is++ ) {
     //    muf2[is] = muf2_hook(is+1);
     //    mur2[is] = mur2_hook(is+1);
     // }
     // std::cout<<fabs( mur2[1] / q2 - 1. )<<std::endl;
     // std::cout<<fabs( muf2[2] / p2 - 1. )<<std::endl;
     // std::cout<<fabs( muf2[3] / p2 - 1. )<<std::endl;
     // std::cout<<fabs( mur2[4] / (p2*p2) - 1. )<<std::endl;
     // std::cout<<fabs( muf2[5] / (q2*q2) - 1. )<<std::endl;
     // std::cout<<fabs( muf2[6] / (q2*q2) - 1. )<<std::endl;
     // const double EPS = 1.e-10;
     // if ( fabs( muf2[1] / q2 - 1 ) > EPS || fabs( mur2[1] / q2 - 1 ) > EPS) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [1]"<< std::endl;
     //    exit(3);
     // }
     // if ( fabs( muf2[2] / p2 - 1 ) > EPS || fabs( mur2[2] / p2 - 1 ) > EPS ) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [2]"<< std::endl;
     //    exit(3);
     // }
     // if ( fabs( muf2[3] / p2 - 1 ) > EPS || fabs( mur2[3] / q2 - 1 ) > EPS) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [3]"<< std::endl;
     //    exit(3);
     // }
     // if ( fabs( muf2[4] / q2 - 1 ) > EPS || fabs( mur2[4] / (p2*p2) - 1 ) > EPS) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [4]"<< std::endl;
     //    exit(3);
     // }
     // if ( fabs( muf2[5] / (q2*q2) - 1 ) > EPS || fabs( mur2[5] / p2 - 1 ) > EPS ) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [5]"<< std::endl;
     //    exit(3);
     // }
     // // TEST
     // if ( fabs( muf2[6] / (q2*q2) - 1 ) > EPS || fabs( mur2[6] / (p2*p2) - 1 ) > EPS ) {
     //    std::cerr<<"[nnlo::fill_fastnlo]. Error. This is a flexible scale table, but the definition of the scales does not agree with the convention [6]"<< std::endl;
     //    exit(3);
     // }
   }
   // For fixed-scale tables
   else {
     // Get list of scale variations requested for fastNLO
     const std::vector<double> ScaleVars = fnloUtils::ftable[id]->GetScaleVariations();
     unsigned int nSclfN = ScaleVars.size();
     for ( unsigned int is=0; is<nSclfN; is++) {
       std::cout << "[fnloUtils::InitFastNLO] Scale variations requested for fastNLO: is = "<<is<<", ScaleVar = "<<ScaleVars[is]<<std::endl;
     }
     // Only one muf scale variation, 1.0, needs to be stored at LO
     if ( nnlo::GetOrder() == 0 ) {
       if ( ScaleVars.size() != 1 || std::abs(ScaleVars[0]-1.0) > DBL_MIN ) {
         std::cerr<<"[fnloUtils::InitFastNLO] ERROR! At LO only one scale factor of 1.0 is necessary. Aborted!"<< std::endl;
         exit(11);
       }
       // The central (first) scale in NNLOJET must have mur=muf for fixed-scale tables
       std::cout<<"[fnloUtils::InitFastNLO] NNLOJET central scale: mur2 = " << mur2_hook(1) << ", muf2 = " << muf2_hook(1) << ", mur2/muf2 = " << mur2_hook(1)/muf2_hook(1) << std::endl;
       if ( std::abs(mur2_hook(1)/muf2_hook(1)-1.0) > DBL_MIN ) {
         std::cerr<<"[fnloUtils::InitFastNLO] ERROR! Fixed-scale tables must have mur=muf for the first scale variation. Aborted!"<< std::endl;
         exit(11);
       }
     } else {
       // Calculate scale factors from NNLOJET SCALES and compare to requested scale variations
       unsigned int nSclNN = 0;
       for (int is = 1; is<maxscl+1; is++) {
         if (std::abs(mur2_hook(is)) > DBL_MIN) { // Filled?
           nSclNN++;
           std::cout<<"[fnloUtils::InitFastNLO] NNLOJET scale var is: " << is << ", mur(is) = " << sqrt(mur2_hook(is)) << ", muf(is) = " << sqrt(mur2_hook(is)) << ", wt(is) = " << wt[is] << std::endl;
         }
       }
       if ( !fnloUtils::ftable[id]->GetIsWarmup() && nSclNN != nSclfN ) { // TODO: More NNLOJET scale vars than for fastNLO (only muf variations need to be stored)
         std::cerr<<"[fnloUtils::InitFastNLO] ERROR! Unequal no. of scale variations requested. Aborted!"<< std::endl;
         std::cerr<<"[fnloUtils::InitFastNLO] NNLOJET: nSclNN = "<<nSclNN<<", fastNLO nScaleVars = "<<nSclfN<< std::endl;
         std::cerr<<"[fnloUtils::InitFastNLO] It is possible to have many NNLOJET variations, but to store in fastNLO only the ones with muf different."<< std::endl;
         std::cerr<<"[fnloUtils::InitFastNLO] This has not yet been implemented. TODO"<< std::endl;
         exit(12);
       }
       for (unsigned int is = 1; is<nSclfN+1; is++) {
         std::cout<<"[fnloUtils::InitFastNLO] NNLOJET scale factors^2 are: " << is << ", mur2(is)/mur2(1) = " << mur2_hook(is)/mur2_hook(1) << ", muf2(is)/muf2(1) = " << muf2_hook(is)/muf2_hook(1) << std::endl;
         std::cout<<"[fnloUtils::InitFastNLO] fastNLO scale factor^2 is: " << is-1 << ", ScaleVar^2 = " << ScaleVars[is-1]*ScaleVars[is-1] << std::endl;
         if ( (std::abs(mur2_hook(is)/mur2_hook(1) - ScaleVars[is-1]*ScaleVars[is-1]) > DBL_MIN) ||
              (std::abs(muf2_hook(is)/muf2_hook(1) - ScaleVars[is-1]*ScaleVars[is-1]) > DBL_MIN ) ) {
           std::cerr<<"[fnloUtils::InitFastNLO] ERROR! NNLOJET and fastNLO scale factors differ, please correct. Aborted!"<< std::endl;
           std::cerr<<"[fnloUtils::InitFastNLO] More flexibility is possible, but has not yet been implemented. TODO"<< std::endl;
           exit(12);
         }
       }
     }
     // OLD CODE
     // // TODO: The scale setting interface is extremely intransparent ==> IMPROVE
     // // --- adapt for muf-variations for fixed-scale tables
     // static const double sclfac[maxscl] = {1,0.5,2.,0.505, 202,0.005,200}; // ugly convention: muf*0.5, muf*2, mur+muf * 0.5, mur+muf* 2, mur*0.5 mur*2
     // for ( int is = 1 ; is<maxscl && fabs(wt[is]) > DBL_MIN; is++ )  {
     //   //         std::cout<<"[fnloUtils::InitFastNLO] wt["<<is<<"]="<<wt[is]<< std::endl;
     //   sc.ScaleVariationFactors.push_back(sclfac[is]);
     // }
     // TODO Improve. Only 0,1,2 of NNLOJET runcard SCALES are accessed.
     // SCALES must be defined to fulfil the following conditions!
     // for ( int is = 0 ; is< 3 && is<(int)sc.ScaleVariationFactors.size() ; is++ ){
     //   // mur(1) is base scale
     //   double mur0  = mur2_hook(1);
     //   //double muf0  = muf2_hook(1);
     //   // This loops over 1,2,3
     //   double mur2  = mur2_hook(is+1);
     //   double muf2  = muf2_hook(is+1);
     //   if ( is==0 && mur2!=muf2 ) { // mur==muf for first "scale variation"
     //          std::cerr<<"[nnlo::fill_fastnlo]. Error. First scale setting: muf and mur must be identical, but muf2/mur2="<<muf2/mur2<< std::endl;
     //          exit(2);
     //   }
     //   // else if ( is==1 && fabs(muf2/mur2/0.25-1) > 1.e-6 ) {
     //   //    std::cerr<<"[nnlo::fill_fastnlo]. Error. Second scale variation must be muf*0.5=mur*0.5, but muf2/mur2="<<muf2/mur2<< std::endl;
     //   //    exit(2);
     //   // }
     //   // else if ( is==2 && fabs(muf2/mur2/4-1) > 1.e-6 ) {
     //   //    std::cerr<<"[nnlo::fill_fastnlo]. Error. Third scale variation must be muf*2=mur*2, but muf2/mur2="<<muf2/mur2<< std::endl;
     //   //    exit(2);
     //   // }
     //   if ( fabs(muf2/mur2-1) > 1.e-6 ) { // mur/muf = 1 for all "scale variations" ==> symmetric variation of muf, mur versus base scale mur(1)
     //          std::cerr<<"[nnlo::fill_fastnlo]. Error. Scale variation is="<<is<<" must be muf*0.5=mur*0.5, but muf2/mur2="<<muf2/mur2<< std::endl;
     //          exit(2);
     //   }
     //   if ( is==1 && fabs(mur2/mur0/0.25-1) > 1.e-6 ) { // mur/mu0 = 2 && muf/mu0 = 2
     //          std::cerr<<"[nnlo::fill_fastnlo]. Error. Second scale variation must be muf*0.5=mur*0.5, but mur2/mur2(0)="<<mur2/mur0<< std::endl;
     //          exit(2);
     //   }
     //   if ( is==2 && fabs(mur2/mur0/4.-1) > 1.e-6 ) { // mur/mu0 = 1/2 && muf/mu0 = 1/2
     //          std::cerr<<"[nnlo::fill_fastnlo]. Error. Second scale variation must be muf*2=mur*2, but mur2/mur2(0)="<<mur2/mur0<< std::endl;
     //          exit(2);
     //   }
     // }
     // //         if ( fabs(wt[3])>TINY ) {
     // // KR: Use C++ macro constants like DBL_MIN instead
     // if ( fabs(wt[3])>DBL_MIN ) {
     //   std::cerr<<"[nnlo::fill_fastnlo]. Error. fastNLO only support scale settings of [0] mur=muf; [1] muf*0.5=mur*0.5; [2] muf*2=muf*2"<< std::endl; // Yes.
     //   exit(2);
     // }
   }

   std::cout << "[fnloUtils::InitFastNLO] fastNLO grid no. "<<id<<" initialized." << std::endl;
   return ;
}



// _____________________________________________________________________ //
namespace fnloUtils {

   // _____________________________________________________________________ //
   fnloSuprocesses::fnloSuprocesses(const std::string& config){
      //!< constructor reading config file with subprocesses
      ReadConfigFile(config);
   }

   // _____________________________________________________________________ //
  void fnloSuprocesses::ReadConfigFile(const std::string& config) {
      fConfigFile = config;
      this->resize(1000);
      fInitialValues.resize(1000);
      std::ifstream cf(config);
      if ( !cf.is_open() ){
         std::cerr<<"[fnloSubprocesses::ReadConfigFile()] Error. Could not open file: "<<config<< std::endl;
         exit(2);
      }
      int id, np, v1, v2;
      int nl = 0;
      cf>>id; // first line is '0'
      while ( !cf.eof() ) {
         cf>>id>>np;
         if(cf.fail()) break;
         nl++;
         //cout<<"nl: "<<nl<<"\tid="<<id<<"\tnp="<<np<< std::endl;
         for ( int i = 0 ; i<np ; i++ ) {
            cf>>v1>>v2;
            //cout<<"\t"<<v1<<"\t"<<v2<< std::endl;
            fInitialValues[id].push_back(v1);
            fInitialValues[id].push_back(v2);
         }
      }
      std::cout<<"[fnloSuprocesses::ReadConfigFile()] found "<<nl<<" subprocess definitions."<< std::endl;

      SetUniqueValues();
   }


   // _____________________________________________________________________ //
   bool fnloSuprocesses::CompareVectors(const std::vector<int>& v1, const std::vector<int>& v2 ) const {
      //!< compare to vectors if all elements are equal
      if ( v1.size() != v2.size() ) return false;
      for ( unsigned int i=0 ; i<v1.size() ; i++ ) {
         if ( v1[i] != v2[i] ) return false;
      }
      return true;
   }


   // _____________________________________________________________________ //
   void fnloSuprocesses::SetUniqueValues() {
      //!< init member SetUniqueValues and 'mother' vector
      for ( unsigned int i = 0 ; i<fInitialValues.size() ; i++ ) {
         if ( !fInitialValues[i].empty() ){
            // look if values are already present ?
            unsigned int umax = fUniqueValues.size();
            for ( unsigned int u = 0 ; u <= umax ; u++ ) {
               if ( u==fUniqueValues.size() ) fUniqueValues.push_back(fInitialValues[i]);
               else if ( CompareVectors(fInitialValues[i],fUniqueValues[u]) )  break;
            }
         }
      }

      // sort a bit ?!
      // put g-g to 0 if available
      vector<int> gg(2);//{0,0}; gg -> X
      for ( unsigned int u = 0 ; u < fUniqueValues.size() ; u++ ) {
         if ( CompareVectors(gg,fUniqueValues[u])) {
            std::swap(fUniqueValues[u],fUniqueValues[0]);
         }
      }
      gg[1]=99;//{0,99}; // DIS eg -> X
      for ( unsigned int u = 0 ; u < fUniqueValues.size() ; u++ ) {
         if ( CompareVectors(gg,fUniqueValues[u])) {
            std::swap(fUniqueValues[u],fUniqueValues[0]);
         }
      }

      // define our map (which is actually a vector)
      for ( unsigned int i = 0 ; i<fInitialValues.size() ; i++ ) {
         (*this)[i] = -1; // init values
         if ( !fInitialValues[i].empty() ){
            for ( unsigned int u = 0 ; u < fUniqueValues.size() ; u++ ) {
               if ( CompareVectors(fInitialValues[i],fUniqueValues[u]) )
                  (*this)[i] = u; // define our final map
            }
         }
      }

      std::cout<<"[fnloSuprocesses::SetUniqueValues()] found "<<fUniqueValues.size()<<" unique subprocess definitions."<<std::endl;
      std::cout<<"[fnloSuprocesses::SetUniqueValues()] Initial values from config file:"<<std::endl;
      for ( unsigned int i = 0 ; i<fInitialValues.size() ; i++ ) if ( fInitialValues[i].size()) std::cout<<i<<"\t"<<fInitialValues[i] << std::endl;
      std::cout<<"[fnloSuprocesses::SetUniqueValues()] Processes definitions:"<<std::endl;
      for ( unsigned int i = 0 ; i<fUniqueValues.size() ; i++ )  std::cout<<i<<"\t"<<fUniqueValues[i] << std::endl;
      std::cout<<"[fnloSuprocesses::SetUniqueValues()] Mapping:"<<std::endl;
      for ( unsigned int i = 0 ; i<fInitialValues.size() ; i++ ) if ( (*this)[i] > -1 ) std::cout<<i<<"\t"<<(*this)[i] << std::endl;
      //      std::cout<<"[fnloSuprocesses::SetUniqueValues()] Mapping: "<<(*this)<< std::endl<< std::endl;
   }

   // _____________________________________________________________________ //
   std::vector<std::vector<int > >  fnloSuprocesses::GetPartonCombinations() const {
      auto out = fUniqueValues;
      int n = 0;
      for (auto& i : out ) i.emplace(i.begin(),n++);
      //for ( auto& i : PartonCombinationsLO ) std::cout<<i<< std::endl;
      return out;
   }

} // end namespace

#endif

