
// emacs: this is -*- c++ -*-
//
//   @file    nnlo_interface.cxx
//
//            grid filling routines for interface with
//            the nnlojet code from CITATION
//
//   Copyright (C) 2015 M.Sutton (sutt@cern.ch) et al.
//
//   $Id: nnlo_interface.h, v0.0  Thu 4 Jul 2015 10:25:18 CEST sutt $


#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <cmath>

#include <sys/stat.h>

#include "amconfig.h"

// nnlo interface header files
#include "nnlo_interface.h"
#include "nnlo_utils.h"
#include "nnlo_common.h"


#ifdef HAVE_APPL
// applgrid headers
#include "appl_grid/lumi_pdf.h"
#include "applgrid_utils.h"
#endif

#ifdef HAVE_FASTNLO
/// fastnlo headers ...
#include "fastnlo_utils.h"
#endif



// external hook functions
extern void (*appl_initptr)( const int& igrid, const int& id, const std::string& gridname,
                             const int& nbins, const double* bins );

extern void (*appl_fillptr)( const int& igrid, const int& id, const double& obs, const double* wt, const double* wtref  );

extern void (*appl_termptr)( const int& igrid, const int& id );

extern "C" void evolvepdf_(const double& x, const double& Q, double* xf);
extern "C" double alphaspdf_(const double& Q);



/// Declare the input and output grids
/// Declare grid-specific functions
static std::string _gridname; //  = "nnlojet.root";


// external hook functions
extern "C" bool setup_nnlo_() {
  std::cout << "[nnlo-bridge] NNLOJET INFO: setting up hook functions ..." << std::endl;

  appl_initptr       = nnlo::init;
  appl_fillptr       = nnlo::fill;
  appl_termptr       = nnlo::term;

  return true;
}



/// Initialise the hook functions
static bool _ready = setup_nnlo_();



// #ifdef HAVE_APPL
// //static std::int  _gridname; //  = "nnlojet.root";
// static std::vector<appl::grid*> _grids;
// static std::map<int, appl::grid*> _gridmap;
// static std::map<int, std::string> _gridnames;

// void nnlo::fill_appl_internal( appl::grid* _grid,
//                                const int& order,
//                                const double& x1, const double& x2, const double& Q2,
//                                const double& obs, const double* wgt );

// appl::grid* find( const int& id ) {
//    std::map<int, appl::grid*>::iterator gitr = _gridmap.find(id);
//    if ( gitr!=_gridmap.end() ) return gitr->second;
//    std::cerr << "find() requested grid id " << id << " not found" << std::endl;
//    return 0;
// }
// #endif


template<typename T>
std::ostream& operator<<( std::ostream& s, std::vector<T>& t ) {
  for ( size_t i=0 ; i<t.size() ; i++ ) s << "\t" << t[i];
  return s;
}



// Temporary pointer to APPLGRID lumi function for fastNLO
//const lumi_pdf *fnlopdf;



bool first_run = true;

//!
//! Initialisation; book interpolation grid(s)
//!
void nnlo::init( const int& igrid,
                 const int& id, const std::string& gridname,
                 const int& nbins, const double* bins ) {

   // --- print welcome message
   static unsigned int call_count = 0;
   call_count++;
   if ( call_count==1 ) {
     std::cout << banner() << std::endl;
     std::cout << "\tinit() " << date() << "\tcompiled: " << __DATE__ << std::endl;
#ifdef HAVE_FASTNLO
     // Check consistent setting of maximal scale variations in NNLOJET and nnlo-bridge
     if ( fnloUtils::GetMaxscl() != maxscl ) {
       std::cerr << "### nnlo::init(): ERROR! Inconsistent setting discovered of maximal scale variations, aborted!" << std::endl;
       std::cerr << "### nnlo::init(): NNLOJET: GetMaxscl() = "<<fnloUtils::GetMaxscl()<<", nnlo_common.h: maxscl = "<<maxscl<< std::endl;
       std::cerr << "### nnlo::init(): Please adapt setting in nnlo_common.h and recompile and relink."<< std::endl;
       exit(9);
     }
#endif
     if ( ! _ready ) {
        std::cout<<"[nnlo::init()]. Warning. Hook-function setup setup_nnlo_() returned false value."<<std::endl;
     }
   }

   // --- cut off grid name extension (appl,fast,both; and at the same time strange additional symbols ...)
   // --- there might be multiple grids to be booked --> multiple init calls!
   _gridname = gridname;
   if ( gridname.find(".fast")!=std::string::npos ) _gridname = _gridname.substr(0,_gridname.find(".fast"));
   else if ( gridname.find(".appl")!=std::string::npos ) _gridname = _gridname.substr(0,_gridname.find(".appl"));
   else if ( gridname.find(".grid")!=std::string::npos ) _gridname = _gridname.substr(0,_gridname.find(".grid"));
   else if ( gridname.find(".both")!=std::string::npos ) _gridname = _gridname.substr(0,_gridname.find(".both"));
   else {
      std::cerr<<"[nnlo::init()]. Error. Cannot identify gridname. Gridname must include '.fast', '.appl', '.grid', or '.both'. Exiting."<<std::endl;
      exit(4);
   }

   std::cout << "### nnlo::init(): Analysing init call ..." << std::endl;
   std::cout << "### nnlo::init(): grid id = " << igrid << " , hist id = " << id << " , gridname = " << _gridname <<std::endl;
   std::cout << "### nnlo::init(): nbins = " << nbins << " , lo = " << bins[0] << " , hi = " << bins[nbins] << std::endl;

   // --- check if grid-tools are available as requested
   // --- init APPLgrid
   if ( igrid==1 || igrid==3 ) { // APPLGRID requested
#    ifdef HAVE_APPL
      std::cout << "### nnlo::init(): init applgrid." << std::endl;
      init_applgrid( id, _gridname, nbins, bins );
#    else
      std::cerr << "### nnlo::init(): APPLGRID requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }

   // --- init fastNLO
   if ( igrid==2 || igrid==3 ) { // fastNLO requested
#    ifdef HAVE_FASTNLO
      std::cout << "### nnlo::init(): init fastNLO." << std::endl;
      init_fastnlo( id, _gridname, nbins, bins );
#    else
      std::cerr << "### nnlo::fill(): fastNLO requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }

   // --- success
   std::cout << "### nnlo::init(): End Init ..." << std::endl;
}



//
// Fill the interpolation grid with the weights
//
void nnlo::fill(  const int& igrid,
                  const int& id, const double& obs, const double* wt, const double* wtref ) {

   static unsigned int call_count = 0;
   call_count++;

   //   std::cout << "nnlo::fill() called " << call_count << " times at " << date() << std::endl;
   //   std::cout << "### nnlo::fill(): Fill grid ..." << std::endl;


   if ( igrid==1 || igrid==3 ) { // APPLGRID requested
#    ifdef HAVE_APPL
   /// nominal kinematics
   /// use common/parfrac/ rather than common/pdf/
   //  double x1 =  pdf_.xi1;
   //  double x2 =  pdf_.xi2;
   double x1 =  parfrac_.x1;
   double x2 =  parfrac_.x2;
   //   double Q2 =  scales_.rscale[0];
   double Q2 =  mur2_hook(1);
   //  double Q2 =  Q*Q;
      fill_applgrid( id, x1, x2, Q2, obs, wt );
      fill_ref_applgrid( id, obs, wtref );
#    else
      std::cerr << "### nnlo::fill(): APPLGRID requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }


   // --- fastNLO requested
   if ( igrid==2 || igrid==3 ) {
#    ifdef HAVE_FASTNLO
      fill_fastnlo( id, obs, wt, wtref );
      fill_ref_fastnlo( id, obs, wtref );
#    else
      std::cerr << "### nnlo::fill(): fastNLO requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }

   //   std::cout << "### nnlo::fill(): End filling ..." << std::endl;
}



// Write out the grids
//
void nnlo::term(  const int& igrid,
                  const int& id ) {

   static unsigned int call_count = 0;
   call_count++;

   std::cout << "### nnlo::term() called " << call_count << " for igrid=" << igrid << " and id=" << id << " times at " << date() << std::endl;
   std::cout << "  iterations: " << runinfo_.itmax2 << "\n"
             << "  shots:      " << runinfo_.nevents << "\n"
             << "  total:      " << runinfo_.itmax2*runinfo_.nevents<<std::endl;

   double nweights = runinfo_.itmax2*runinfo_.nevents;

   if ( igrid==1 || igrid==3 ) { // APPLGRID requested
#    ifdef HAVE_APPL
      term_applgrid( id, nweights );
#    else
      std::cerr << "### nnlo::term(): APPLGRID requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }

   // --- fastNLO requested
   if ( igrid==2 || igrid==3 ) {
#    ifdef HAVE_FASTNLO
      term_fastnlo( id, nweights );
#    else
      std::cerr << "### nnlo::term(): fastNLO requested, but not linked!" << std::endl;
      std::exit(1);
#    endif
   }

   std::cout << "### nnlo::term(): End Term ..." << std::endl;
}

